from syned.beamline.shape import SurfaceShape
from syned.beamline.optical_element_with_surface_shape import OpticalElementsWithSurfaceShape

class DiffractionGeometry:
    BRAGG = 0
    LAUE = 1

class Crystal(OpticalElementsWithSurfaceShape):
    def __init__(self,
                 name="Undefined",
                 surface_shape=SurfaceShape(),
                 boundary_shape=None,
                 material="Si",
                 diffraction_geometry=DiffractionGeometry.BRAGG,
                 miller_index_h=1,
                 miller_index_k=1,
                 miller_index_l=1,
                 asymmetry_angle=0.0,
                 thickness=0.0,
                 ):

        super().__init__(name, surface_shape, boundary_shape)
        self._material = material
        self._diffraction_geometry = diffraction_geometry
        self._miller_index_h = miller_index_h
        self._miller_index_k = miller_index_k
        self._miller_index_l = miller_index_l
        self._asymmetry_angle = asymmetry_angle
        self._thickness = thickness

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                "Name" ,                  "" ),
                    ("surface_shape",       "Surface Shape" ,         "" ),
                    ("boundary_shape",      "Boundary Shape" ,        "" ),
                    ("material",            "Material (name)" ,       "" ),
                    ("diffraction_geometry","Diffraction Geometry",   "" ),
                    ("miller_index_h",      "Miller index h",         "" ),
                    ("miller_index_k",      "Miller index k",         "" ),
                    ("miller_index_l",      "Miller index l",         "" ),
                    ("asymmetry_angle",     "Asymmetry angle",        "rad"),
                    ("thickness",           "Thickness",              "m"),
            ] )





from syned.beamline.beamline import Beamline

class WidgetDecorator(object):

    @classmethod
    def syned_input_data(cls):
        return [("SynedData", Beamline, "receive_syned_data")]

    @classmethod
    def append_syned_input_data(cls, inputs):
        for input in WidgetDecorator.syned_input_data():
            inputs.append(input)

    def receive_syned_data(self, data):
        raise NotImplementedError("Should be implemented")


"""

Base class for all insertion devices: wiggler, undulator

"""

from numpy import pi
import scipy.constants as codata

from syned.storage_ring.magnetic_structure import MagneticStructure

class InsertionDevice(MagneticStructure):
    def __init__(self,
                 K_vertical        = 0.0,
                 K_horizontal      = 0.0,
                 period_length     = 0.0,
                 number_of_periods = 1.0):
        MagneticStructure.__init__(self)

        self._K_vertical = K_vertical
        self._K_horizontal = K_horizontal
        self._period_length = period_length
        self._number_of_periods = number_of_periods

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("K_vertical"          , "K value (vertical)"  , ""    ),
                    ("K_horizontal"        , "K value (horizontal)", ""    ),
                    ("period_length"       , "Period length"       , "m"   ),
                    ("number_of_periods"   , "Number of periods"   , ""    ),
            ] )

    def K_vertical(self):
        return self._K_vertical

    def K_horizontal(self):
        return self._K_horizontal

    def period_length(self):
        return self._period_length

    def number_of_periods(self):
        return self._number_of_periods


    #
    # some easy calculations
    #

    def K(self):
        return self.K_vertical()

    def length(self):
        return self.number_of_periods() * self.period_length()

    def _magnetic_field_from_K(self, K):
        return K * 2 * pi * codata.m_e * codata.c / (codata.e * self.period_length())

    def _K_from_magnetic_field(self, B):
        return B /(2 * pi * codata.m_e * codata.c / (codata.e * self.period_length()))

    def magnetic_field_vertical(self):
        return self._magnetic_field_from_K(self.K_vertical())

    def magnetic_field_horizontal(self):
        return self._magnetic_field_from_K(self.K_horizontal())

    def set_K_vertical_from_magnetic_field(self, B_vertical):
        self._K_vertical = self._K_from_magnetic_field(B_vertical)

    def set_K_horizontal_from_magnetic_field(self, B_horizontal):
        self._K_horizontal = self._K_from_magnetic_field(B_horizontal)


"""

Represents an ideal lens.

"""
from syned.beamline.optical_elements.ideal_elements.ideal_element import IdealElement


class IdealLens(IdealElement):
    def __init__(self, name="Undefined", focal_x=1.0, focal_y=1.0):
        IdealElement.__init__(self, name=name)
        self._focal_x = focal_x
        self._focal_y = focal_y
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("focal_x"      , "Focal length in x [horizontal]", "m" ),
                    ("focal_y"      , "Focal length in y [vertical]",    "m" ),
            ] )

    def focal_x(self):
        return self._focal_x

    def focal_y(self):
        return self._focal_y



from syned.beamline.shape import SurfaceShape, BoundaryShape
from syned.beamline.optical_element_with_surface_shape import OpticalElementsWithSurfaceShape

class Grating(OpticalElementsWithSurfaceShape):
    def __init__(self,
                 name="Undefined",
                 surface_shape=SurfaceShape(),
                 boundary_shape=BoundaryShape(),
                 ruling=800e3,
                 ):
        super().__init__(name, surface_shape, boundary_shape)
        self._ruling = ruling

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                   "Name" ,                  "" ),
                    ("surface_shape",          "Surface Shape" ,         "" ),
                    ("boundary_shape",         "Boundary Shape" ,        "" ),
                    ("ruling",                 "Ruling at center" ,      "lines/m" ),
            ] )


class GratingVLS(Grating):
    def __init__(self,
                 name="Undefined",
                 surface_shape=SurfaceShape(),
                 boundary_shape=BoundaryShape(),
                 ruling=800e3,
                 ruling_coeff_linear=0.0,
                 ruling_coeff_quadratic=0.0,
                 ruling_coeff_cubic=0.0,
                 ruling_coeff_quartic=0.0,
                 coating=None,
                 coating_thickness=None,
                 ):
        super().__init__(name, surface_shape, boundary_shape)

        self._ruling = ruling
        self._ruling_coeff_linear = ruling_coeff_linear
        self._ruling_coeff_quadratic = ruling_coeff_quadratic
        self._ruling_coeff_cubic = ruling_coeff_cubic
        self._ruling_coeff_quartic = ruling_coeff_quartic
        self._coating = coating
        self._coating_thickness = coating_thickness

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                   "Name" ,                  "" ),
                    ("surface_shape",          "Surface Shape" ,         "" ),
                    ("boundary_shape",         "Boundary Shape" ,        "" ),
                    ("ruling",                 "Ruling at center" ,      "lines/m" ),
                    ("ruling_coeff_linear",    "Ruling linear coeff",    "lines/m^2"),
                    ("ruling_coeff_quadratic", "Ruling quadratic coeff", "lines/m^3"),
                    ("ruling_coeff_cubic",     "Ruling cubic coeff",     "lines/m^4"),
                    ("ruling_coeff_quartic",   "Ruling quartic coeff",    "lines/m^5"),
                    ("coating",                "Coating (element, compound or name)", ""),
                    ("coating_thickness",      "Coating thickness", "m"),
            ] )


class GratingBlaze(GratingVLS):
    def __init__(self,
                 name="Undefined",
                 surface_shape=SurfaceShape(),
                 boundary_shape=BoundaryShape(),
                 ruling=800e3,
                 ruling_coeff_linear=0.0,
                 ruling_coeff_quadratic=0.0,
                 ruling_coeff_cubic=0.0,
                 ruling_coeff_quartic=0.0,
                 coating=None,
                 coating_thickness=None,
                 blaze_angle=0.0,
                 antiblaze_angle=90.0,
                 ):
        super().__init__(name, surface_shape, boundary_shape,
                         ruling=ruling,
                         ruling_coeff_linear=ruling_coeff_linear,
                         ruling_coeff_quadratic=ruling_coeff_quadratic,
                         ruling_coeff_cubic=ruling_coeff_cubic,
                         ruling_coeff_quartic=ruling_coeff_quartic,
                         coating=coating,
                         coating_thickness=coating_thickness,
                         )
        self._blaze_angle = blaze_angle
        self._antiblaze_angle = antiblaze_angle


        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name", "Name", ""),
                    ("surface_shape", "Surface Shape", ""),
                    ("boundary_shape", "Boundary Shape", ""),
                    ("ruling", "Ruling at center", "lines/m"),
                    ("ruling_coeff_linear", "Ruling linear coeff", "lines/m^2"),
                    ("ruling_coeff_quadratic", "Ruling quadratic coeff", "lines/m^3"),
                    ("ruling_coeff_cubic", "Ruling cubic coeff", "lines/m^4"),
                    ("ruling_coeff_quartic", "Ruling quartic coeff", "lines/m^5"),
                    ("coating", "Coating (element, compound or name)", ""),
                    ("coating_thickness", "Coating thickness", "m"),
                    ("blaze_angle",      "Blaze angle",     "rad"),
                    ("antiblaze_angle",  "Antiblaze angle", "rad"),
            ] )

    def get_apex_angle(self):
        return 180 - self._blaze_angle - self._antiblaze_angle


class GratingLamellar(GratingVLS):
    def __init__(self,
                 name="Undefined",
                 surface_shape=SurfaceShape(),
                 boundary_shape=BoundaryShape(),
                 ruling=800e3,
                 ruling_coeff_linear=0.0,
                 ruling_coeff_quadratic=0.0,
                 ruling_coeff_cubic=0.0,
                 coating=None,
                 coating_thickness=None,
                 height=1e-6,
                 ratio_valley_to_period=0.5,
                 ):
        super().__init__(name, surface_shape, boundary_shape,
                         ruling=ruling,
                         ruling_coeff_linear=ruling_coeff_linear,
                         ruling_coeff_quadratic=ruling_coeff_quadratic,
                         ruling_coeff_cubic=ruling_coeff_cubic,
                         coating=coating,
                         coating_thickness=coating_thickness,
                         )
        self._height = height
        self._ratio_valley_to_period = ratio_valley_to_period


        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name", "Name", ""),
                    ("surface_shape", "Surface Shape", ""),
                    ("boundary_shape", "Boundary Shape", ""),
                    ("ruling", "Ruling at center", "lines/m"),
                    ("ruling_coeff_linear", "Ruling linear coeff", "lines/m^2"),
                    ("ruling_coeff_quadratic", "Ruling quadratic coeff", "lines/m^3"),
                    ("ruling_coeff_cubic", "Ruling cubic coeff", "lines/m^4"),
                    ("ruling_coeff_quartic", "Ruling quartic coeff", "lines/m^5"),
                    ("coating", "Coating (element, compound or name)", ""),
                    ("coating_thickness", "Coating thickness", "m"),
                    ("height",            "Height",     "m"),
                    ("ratio_valley_to_period",  "Valley/period ratio", ""),
            ] )


if __name__ == "__main__":

    grating1 = Grating(name="grating1")
    # grating1.keys()
    print(grating1.info())
    # print(grating1.to_json())

    grating1 = GratingVLS(name="grating1")
    # grating1.keys()
    print(grating1.info())
    # print(grating1.to_json())

    grating1 = GratingBlaze(name="grating1")
    # grating1.keys()
    print(grating1.info())
    # print(grating1.to_json())

    grating1 = GratingLamellar(name="grating1")
    # grating1.keys()
    print(grating1.info())
    # print(grating1.to_json())






"""

Base class for a  wiggler.

"""

from syned.storage_ring.magnetic_structures.insertion_device import InsertionDevice

class Wiggler(InsertionDevice):

    def __init__(self, K_vertical = 0.0, K_horizontal = 0.0,period_length = 0.0, number_of_periods = 1):
        InsertionDevice.__init__(self,
                                 K_vertical=K_vertical,
                                 K_horizontal=K_horizontal,
                                 period_length=period_length,
                                 number_of_periods=number_of_periods)





"""
Base class for electron beams.

This class is intentionally shorten for simplicity.
Usually we would need to consider also the electron distribution within the beam.
"""

from syned.syned_object import SynedObject
import scipy.constants as codata
import numpy

class ElectronBeam(SynedObject):
    def __init__(self,
                 energy_in_GeV = 1.0,
                 energy_spread = 0.0,
                 current = 0.1,
                 number_of_bunches = 400,
                 moment_xx=0.0,
                 moment_xxp=0.0,
                 moment_xpxp=0.0,
                 moment_yy=0.0,
                 moment_yyp=0.0,
                 moment_ypyp=0.0):


        self._energy_in_GeV       = energy_in_GeV
        self._energy_spread       = energy_spread
        self._current             = current
        self._number_of_bunches   = number_of_bunches

        self._moment_xx           = moment_xx
        self._moment_xxp          = moment_xxp
        self._moment_xpxp         = moment_xpxp
        self._moment_yy           = moment_yy
        self._moment_yyp          = moment_yyp
        self._moment_ypyp         = moment_ypyp

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("energy_in_GeV"      , "Electron beam energy"                  , "GeV" ),
                    ("energy_spread"      , "Electron beam energy spread (relative)", ""    ),
                    ("current"            , "Electron beam current"                 , "A"   ),
                    ("number_of_bunches"  , "Number of bunches"                     , ""    ),
                    ("moment_xx"          , "Moment (spatial^2, horizontal)"        , "m^2" ),
                    ("moment_xxp"         , "Moment (spatial-angular, horizontal)"  , "m"   ),
                    ("moment_xpxp"        , "Moment (angular^2, horizontal)"        , ""    ),
                    ("moment_yy"          , "Moment (spatial^2, vertical)"          , "m^2" ),
                    ("moment_yyp"         , "Moment (spatial-angular, vertical)"    , "m"   ),
                    ("moment_ypyp"        , "Moment (angular^2, vertical)"          , ""    ),
            ] )



    #
    # initializares
    #
    @classmethod
    def initialize_as_pencil_beam(cls, energy_in_GeV = 1.0, energy_spread = 0.0, current = 0.1):
        return ElectronBeam(energy_in_GeV=energy_in_GeV,
                            energy_spread=energy_spread,
                            current=current,
                            number_of_bunches=1)


    #
    # useful getters
    #
    def get_sigmas_real_space(self):
        return numpy.sqrt(self._moment_xx),\
               numpy.sqrt(self._moment_yy)

    def get_sigmas_divergence_space(self):
        return numpy.sqrt(self._moment_xpxp),\
               numpy.sqrt(self._moment_ypyp)

    def get_sigmas_horizontal(self):
        return numpy.sqrt(self._moment_xx),\
               numpy.sqrt(self._moment_xpxp)

    def get_sigmas_vertical(self):
        return numpy.sqrt(self._moment_yy),\
               numpy.sqrt(self._moment_ypyp)

    def get_sigmas_all(self):
        return numpy.sqrt(self._moment_xx),\
               numpy.sqrt(self._moment_xpxp),\
               numpy.sqrt(self._moment_yy),\
               numpy.sqrt(self._moment_ypyp)

    def get_moments_horizontal(self):
        return self._moment_xx,self._moment_xxp,self._moment_xpxp

    def get_moments_vertical(self):
        return self._moment_yy,self._moment_yyp,self._moment_ypyp

    def get_moments_all(self):
        return self._moment_xx,self._moment_xxp,self._moment_xpxp,self._moment_yy,self._moment_yyp,self._moment_ypyp

    def get_twiss_no_dispersion_horizontal(self):
        emittance_x = numpy.sqrt(self._moment_xx * self._moment_xpxp - self._moment_xxp**2)
        alpha_x = -self._moment_xxp / emittance_x
        beta_x = self._moment_xx / emittance_x
        return emittance_x,alpha_x,beta_x

    def get_twiss_no_dispersion_vertical(self):
        emittance_y = numpy.sqrt(self._moment_yy * self._moment_ypyp - self._moment_yyp**2)
        alpha_y = -self._moment_yyp / emittance_y
        beta_y = self._moment_yy / emittance_y
        return emittance_y,alpha_y,beta_y

    def get_twiss_no_dispersion_all(self):
        ex, ax, bx =  self.get_twiss_no_dispersion_horizontal()
        ey, ay, by = self.get_twiss_no_dispersion_vertical()
        return ex, ax, bx, ey, ay, by

    def energy(self):
        return self._energy_in_GeV

    def current(self):
        return self._current

    #
    # setters
    #
    def set_sigmas_real_space(self,sigma_x=0.0,sigma_y=0.0):
        self._moment_xx = sigma_x**2
        self._moment_yy = sigma_y**2

    def set_sigmas_divergence_space(self,sigma_xp=0.0,sigma_yp=0.0):
        self._moment_xpxp = sigma_xp**2
        self._moment_ypyp = sigma_yp**2

    def set_sigmas_horizontal(self,sigma_x=0.0,sigma_xp=0.0):
        self._moment_xx = sigma_x**2
        self._moment_xpxp = sigma_xp**2

    def set_sigmas_vertical(self,sigma_y=0.0,sigma_yp=0.0):
        self._moment_yy = sigma_y**2
        self._moment_ypyp = sigma_yp**2

    def set_sigmas_all(self,sigma_x=0.0,sigma_xp=0.0,sigma_y=0.0,sigma_yp=0.0):
        self._moment_xx = sigma_x**2
        self._moment_xpxp = sigma_xp**2
        self._moment_yy = sigma_y**2
        self._moment_ypyp = sigma_yp**2

    def set_energy_from_gamma(self, gamma):
        self._energy_in_GeV = (gamma / 1e9) * (codata.m_e *  codata.c**2 / codata.e)

    def set_moments_horizontal(self,moment_xx,moment_xxp,moment_xpxp):
        self._moment_xx   = moment_xx
        self._moment_xxp  = moment_xxp
        self._moment_xpxp = moment_xpxp

    def set_moments_vertical(self,moment_yy,moment_yyp,moment_ypyp):
        self._moment_yy   = moment_yy
        self._moment_yyp  = moment_yyp
        self._moment_ypyp = moment_ypyp

    def set_moments_all(self,moment_xx,moment_xxp,moment_xpxp,moment_yy,moment_yyp,moment_ypyp):
        self.set_moments_horizontal(moment_xx,moment_xxp,moment_xpxp)
        self.set_moments_vertical(moment_yy,moment_yyp,moment_ypyp)

    def set_twiss_horizontal(self,emittance_x,alpha_x,beta_x,eta_x=0,etap_x=0):
        gamma_x = (1 + alpha_x**2) / beta_x
        self._moment_xx   = beta_x   * emittance_x + eta_x**2 * self._energy_spread**2
        self._moment_xxp  = -alpha_x * emittance_x + eta_x * etap_x * self._energy_spread ** 2
        self._moment_xpxp = gamma_x  * emittance_x + etap_x**2 * self._energy_spread ** 2

    def set_twiss_vertical(self,emittance_y,alpha_y,beta_y,eta_y=0,etap_y=0):
        gamma_y = (1 + alpha_y**2) / beta_y
        self._moment_yy   = beta_y   * emittance_y + eta_y**2 * self._energy_spread**2
        self._moment_yyp  = -alpha_y * emittance_y + eta_y * etap_y * self._energy_spread ** 2
        self._moment_ypyp = gamma_y  * emittance_y + etap_y**2 * self._energy_spread ** 2

    def set_twiss_all(self,emittance_x,alpha_x,beta_x,eta_x,etap_x,
                           emittance_y,alpha_y,beta_y,eta_y,etap_y):
        self.set_twiss_horizontal(emittance_x,alpha_x,beta_x,eta_x,etap_x)
        self.set_twiss_vertical(emittance_y,alpha_y,beta_y,eta_y,etap_y)

    #
    # some easy calculations
    #
    def gamma(self):
        return self.lorentz_factor()

    def lorentz_factor(self):
        return 1e9*self._energy_in_GeV / (codata.m_e *  codata.c**2 / codata.e)

    def electron_speed(self):
        return numpy.sqrt(1.0 - 1.0 / self.lorentz_factor() ** 2)



    #
    #dictionnary interface, info etc
    #

if __name__ == "__main__":

    a = ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=2.0,current=0.5, energy_spread=0.00095)

    a.set_twiss_horizontal(70e-12, 0.827, 0.34, 0, 0 ) #0.0031, -0.06)
    a.set_twiss_vertical(  70e-12, -10.7, 24.26, 0, 0.0)

    print("twiss data: 70e-12, 0.827, 0.34, 70e-12, -10.7, 24.26,")
    print("sigmas: ",a.get_sigmas_all())
    print("moments: ",a.get_moments_all())
    print("twiss: ",a.get_twiss_no_dispersion_all())


    s = a.get_sigmas_all()
    a.set_sigmas_all(s[0],s[1],s[2],s[3])

    print("\n\nsigmas: ",a.get_sigmas_all())
    print("moments: ",a.get_moments_all())
    print("twiss: ",a.get_twiss_no_dispersion_all())


#    a.to_dictionary()


    # fd = a.to_full_dictionary()
    # dict = a.to_dictionary()
    #
    # print(dict)
    #
    # for key in fd:
    #     print(key,fd[key][0])
    #
    # for key in fd:
    #     print(key,dict[key])
    #
    # print(a.keys())
    # print(a.info())




from syned.beamline.optical_element import OpticalElement
from syned.beamline.shape import BoundaryShape

class Absorber(OpticalElement):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        OpticalElement.__init__(self, name=name, boundary_shape=boundary_shape)


"""
Base clase for magnetic structures (from where BM, Wiggler and Undulator will heritate)

"""

from syned.syned_object import SynedObject

class MagneticStructure(SynedObject):
    def __init__(self):
        super().__init__()



from syned.beamline.optical_element import OpticalElement
from syned.beamline.shape import SurfaceShape



class OpticalElementsWithSurfaceShape(OpticalElement):

    def __init__(self, name, surface_shape=SurfaceShape(), boundary_shape=None, ):
        super().__init__(name, boundary_shape)
        self._surface_shape = surface_shape

    def get_surface_shape(self):
        return self._surface_shape

    def set_surface_shape(self,surface_shape=SurfaceShape()):
        self._surface_shape = surface_shape


from syned.syned_object import SynedObject

class OpticalElement(SynedObject):
    def __init__(self, name="Undefined", boundary_shape=None):
        self._name = name
        self._boundary_shape = boundary_shape

        # TODO check name
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    # ("name"      ,           "to define ", "" ),
                    ("boundary_shape"      , "to define ", "" ),
            ] )

    def get_name(self):
        return self._name

    def get_boundary_shape(self):
        return self._boundary_shape




from syned.beamline.shape import BoundaryShape
from syned.beamline.shape import Rectangle, Ellipse, Circle

from syned.beamline.optical_elements.absorbers.absorber import Absorber

class Slit(Absorber):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Absorber.__init__(self, name=name, boundary_shape=boundary_shape)

    def set_rectangle(self,width=3e-3,height=4e-3,center_x=0.0,center_y=0.0):
        self._boundary_shape=Rectangle(-0.5*width+center_x,0.5*width+center_x,-0.5*height+center_y,0.5*height+center_y)

    def set_circle(self,radius=3e-3,center_x=0.0,center_y=0.0):
        self._boundary_shape=Circle(radius,center_x,center_y)

    def set_ellipse(self,width=3e-3,height=4e-3,center_x=0.0,center_y=0.0):
        self._boundary_shape=Ellipse(-0.5*width+center_x,0.5*width+center_x,-0.5*height+center_y,0.5*height+center_y)


"""
Position of a beamline component within a beamline.
"""

from syned.syned_object import SynedObject

class ElementCoordinates(SynedObject):
    def __init__(self, p = 0.0, q = 0.0, angle_radial=0.0, angle_azimuthal=0.0):
        """

        :param p: distance from previous element.
        :param q: distance to next element.
        :param angle_radial: Radial inclination angle.
        :param angle_azimuthal: Azimuthal inclination angle.
        :return:
        """
        self._p               = p
        self._q               = q
        self._angle_radial    = angle_radial
        self._angle_azimuthal = angle_azimuthal

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("p",                "distance from previous continuation plane", "m"    ),
                    ("q",                "distance to next continuation plane",      "m"    ),
                    ("angle_radial",     "incident angle [to normal]",               "rad"  ),
                    ("angle_azimuthal",  "rotation along beam axis",                 "rad"  ),
                ])

    def p(self):
        return self._p

    def q(self):
        return self._q

    def angle_radial(self):
        return self._angle_radial

    def angle_azimuthal(self):
        return self._angle_azimuthal



from syned.beamline.optical_elements.absorbers.absorber import Absorber

class Filter(Absorber):
    def __init__(self,
                 name="Undefined",
                 material="Be",
                 thickness=1e-3):
        Absorber.__init__(self, name=name)
        self._material = material
        self._thickness = thickness


        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("material"      , "Material (symbol, formula or name)",    "" ),
                    ("thickness"     , "Thickness ",                            "m" ),
            ] )
    def get_material(self):
        return self._material

    def get_thickness(self):
        return self._thickness



from collections import OrderedDict
import json

# TODO: although basic functionality is implemented, the use of exec should be replace by introspection tools
class SynedObject(object):

    def _set_support_text(self,text):
        ordered_support_dict = OrderedDict()
        for e in text:
            ordered_support_dict[e[0]] = (e[1], e[2])
        self._support_dictionary = ordered_support_dict


    #
    # this is the common interface to allow json file input/output and info mechanism
    #
    # standard methods are:
    #   keys
    #   to_dictionary
    #   to_full_dictionary
    #   to_json
    #   info
    #   set_value_from_key_name
    #   get_value_from_key_name
    #
    def keys(self):
        try:
            return self._support_dictionary.keys()
        except:
            return None

    def to_dictionary(self):
        dict_to_save = OrderedDict()
        dict_to_save.update({"CLASS_NAME":self.__class__.__name__})
        try:
            for key in self.keys():
                tmp1 = eval("self._%s" % (key) )
                if isinstance(tmp1,SynedObject):
                    dict_to_save[key] = tmp1.to_dictionary()
                else:
                    dict_to_save[key] = tmp1
        except:
            pass

        return dict_to_save

    def to_full_dictionary(self):
        dict_to_save = OrderedDict()
        dict_to_save.update({"CLASS_NAME":self.__class__.__name__})
        try:
            for key in self.keys():
                tmp1 = eval("self._%s" % (key) )
                if isinstance(tmp1,SynedObject):
                    dict_to_save[key] = tmp1.to_full_dictionary()
                else:
                    mylist = []
                    mylist.append(tmp1)
                    mylist.append(self._support_dictionary[key])
                    dict_to_save[key] = mylist # [tmp1,self._support_dictionary[key]]
        except:
            pass

        return dict_to_save

    def to_json(self,file_name=None):
        dict1 = OrderedDict()
        dict1.update(self.to_dictionary())

        jsn1 = json.dumps(dict1, indent=4, separators=(',', ': '))
        if file_name != None:
            f = open(file_name,'w')
            f.write(jsn1)
            f.close()
            print("File written to disk: %s"%(file_name))
        return jsn1


    def info_recurrent(self,fd,prefix="    "):
        text = ""
        for key in fd.keys():
            if isinstance(fd[key],OrderedDict):
                text += prefix + self.info_recurrent(fd[key])
            elif isinstance(fd[key],str):
                text += prefix + "-------%s---------\n"%fd[key]
            elif isinstance(fd[key],list):
                if isinstance(fd[key][0],OrderedDict):
                    for element in fd[key]:
                        text += self.info_recurrent(element,prefix="    ")
                elif isinstance(fd[key][0],list):
                    for i,element in enumerate(fd[key][0]):
                        text += element.info()
                else:
                    text += prefix + '    %s: %s %s # %s\n' %(key,  repr(fd[key][0]), fd[key][1][1], fd[key][1][0])
            else:
                pass
        return text

    # TODO: not working correctly for beamline
    def info(self):
        return self.info_recurrent( self.to_full_dictionary() )


    def set_value_from_key_name(self,key,value):
        if key in self.keys():
            try:
                exec("self._%s = value" % (key))
                # print("Set variable %s to value: "%key + repr(value))
            except:
                raise ValueError("Cannot set variable %s to value: "%key + repr(value) )
        else:
            raise ValueError("Key %s not accepted by class %s"%(key,self.__class__.__name__))


    def get_value_from_key_name(self,key):
        try:
            value = eval("self._%s" % (key))
            return value
        except:
            raise ValueError("Cannot get variable %s: "%key)






from syned.beamline.shape import SurfaceShape
from syned.beamline.optical_element_with_surface_shape import OpticalElementsWithSurfaceShape

class Mirror(OpticalElementsWithSurfaceShape):
    def __init__(self,
                 name="Undefined",
                 surface_shape=SurfaceShape(),
                 boundary_shape=None,
                 coating=None,
                 coating_thickness=None):

        super().__init__(name, surface_shape, boundary_shape)
        self._coating = coating
        self._coating_thickness = coating_thickness
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                "Name" ,                               "" ),
                    ("surface_shape",       "Surface shape",                       "" ),
                    ("boundary_shape",      "Boundary shape",                      "" ),
                    ("coating",             "Coating (element, compound or name)", "" ),
                    ("coating_thickness",   "Coating thickness",                   "m"),
            ] )





#
# note that even the elements are not explicitely used in the code that follows, they MUST be imported
# here in order to define the objects at run time.
#
from syned.storage_ring.electron_beam import ElectronBeam
from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.storage_ring.magnetic_structures.wiggler import Wiggler
from syned.storage_ring.magnetic_structures.bending_magnet import BendingMagnet
from syned.storage_ring.magnetic_structure import MagneticStructure
from syned.beamline.optical_elements.ideal_elements.screen import Screen
from syned.beamline.optical_elements.ideal_elements.lens import IdealLens
from syned.beamline.optical_elements.absorbers.filter import Filter
from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.optical_elements.absorbers.beam_stopper import BeamStopper
from syned.beamline.optical_elements.mirrors.mirror import Mirror
from syned.beamline.optical_elements.crystals.crystal import Crystal
from syned.beamline.optical_elements.gratings.grating import Grating
from syned.beamline.optical_elements.gratings.grating import GratingBlaze
from syned.beamline.optical_elements.gratings.grating import GratingLamellar

from syned.beamline.shape import BoundaryShape
from syned.beamline.shape import Rectangle, Ellipse, Circle
from syned.beamline.shape import MultiplePatch
from syned.beamline.shape import DoubleRectangle, DoubleEllipse, DoubleCircle

from syned.beamline.shape import SurfaceShape
from syned.beamline.shape import Conic, Plane, Sphere, SphericalCylinder
from syned.beamline.shape import Ellipsoid, EllipticalCylinder
from syned.beamline.shape import Paraboloid, ParabolicCylinder
from syned.beamline.shape import Hyperboloid, HyperbolicCylinder
from syned.beamline.shape import Toroidal, NumbericalMesh



from syned.storage_ring.light_source import LightSource
from syned.storage_ring.empty_light_source import EmptyLightSource

from syned.beamline.beamline import Beamline
from syned.beamline.beamline_element import BeamlineElement
from syned.beamline.element_coordinates import ElementCoordinates

from collections import OrderedDict

import json
from urllib.request import urlopen

#
# this allows to load wofry-elements (e.g. WOSlit) in addition to syned elements (Slit)
# The commented items are not yet implemented
#
try:
    from wofry.beamline.optical_elements.ideal_elements.screen import WOScreen
    from wofry.beamline.optical_elements.ideal_elements.lens import WOIdealLens
    # from wofry.beamline.optical_elements.absorbers.filter import WOFilter
    from wofry.beamline.optical_elements.absorbers.slit import WOSlit
    from wofry.beamline.optical_elements.absorbers.beam_stopper import WOBeamStopper
    # from wofry.beamline.optical_elements.mirrors.mirror import WOMirror
    # from wofry.beamline.optical_elements.crystals.crystal import WOCrystal
    # from wofry.beamline.optical_elements.gratings.grating import WOGrating
except:
    pass




def load_from_json_file(file_name):
    f = open(file_name)
    text = f.read()
    f.close()
    return load_from_json_text(text)

def load_from_json_url(file_url):
    u = urlopen(file_url)
    ur = u.read()
    url = ur.decode(encoding='UTF-8')
    return load_from_json_text(url)


def load_from_json_text(text):
    return load_from_json_dictionary_recurrent(json.loads(text))


def load_from_json_dictionary_recurrent(jsn,verbose=False):

    if verbose: print(jsn.keys())
    if "CLASS_NAME" in jsn.keys():
        if verbose: print("FOUND CLASS NAME: ",jsn["CLASS_NAME"])

        try:
            tmp1 = eval(jsn["CLASS_NAME"]+"()")
            if verbose: print(">>>>",jsn["CLASS_NAME"],type(tmp1))
        except:
            raise RuntimeError("Error evaluating: "+jsn["CLASS_NAME"]+"()")


        if tmp1.keys() is not None:

            for key in tmp1.keys():
                if verbose: print(">>>>processing",key ,type(jsn[key]))
                if isinstance(jsn[key],dict):
                    if verbose: print(">>>>>>>>dictionary found, starting recurrency",key ,type(jsn[key]))
                    tmp2 = load_from_json_dictionary_recurrent(jsn[key])
                    if verbose: print(">>>>2",key,type(tmp2))
                    tmp1.set_value_from_key_name(key,tmp2)
                elif isinstance(jsn[key],list):
                    if verbose: print(">>>>>>>>LIST found, starting recurrency",key ,type(jsn[key]))
                    out_list_of_objects = []
                    for element in jsn[key]:
                        if isinstance(element,dict):
                            if verbose: print(">>>>>>>>LIST found, starting recurrency",key ,type(element))
                            tmp3 = load_from_json_dictionary_recurrent(element)
                            if verbose: print(">>>>3",type(tmp3))
                            out_list_of_objects.append(tmp3)
                    if verbose: print("kkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk",out_list_of_objects)
                    tmp1.set_value_from_key_name(key,out_list_of_objects)
                        # tmp1.set_value_from_key_name(key,tmp2)
                else:
                    if verbose: print(">>>>>>> settingng value for key: ",key," to: ",repr(jsn[key]))
                    tmp1.set_value_from_key_name(key,jsn[key])

        return tmp1


if __name__ == "__main__":

    # file_url = "http://ftp.esrf.eu/pub/scisoft/syned/lightsources/ESRF_ID01_EBS_ppu27_11.json"
    # syned_obj = load_from_json_url(file_url)
    # print(syned_obj.info())


    #
    # file_url = "/home/manuel/Oasys/ALSU-IR-BM.json"
    # syned_obj = load_from_json_file(file_url)
    # print(syned_obj.info())

    file_url = "/home/manuel/Oasys/bl.json"
    syned_obj = load_from_json_file(file_url)
    print(syned_obj.info())




from syned.storage_ring.electron_beam import ElectronBeam
from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.storage_ring.light_source import LightSource
from syned.beamline.shape import *

from syned.beamline.optical_elements.ideal_elements.screen import Screen
from syned.beamline.optical_elements.ideal_elements.lens import IdealLens
from syned.beamline.optical_elements.absorbers.filter import Filter
from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.optical_elements.absorbers.beam_stopper import BeamStopper

from syned.beamline.optical_elements.mirrors.mirror import Mirror
from syned.beamline.optical_elements.crystals.crystal import Crystal
from syned.beamline.optical_elements.gratings.grating import Grating

from syned.beamline.beamline import BeamlineElement, Beamline
from syned.beamline.element_coordinates import ElementCoordinates

if __name__ == "__main__":


    print("==================== LightSource: ==================")


    src1 = ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=6.0,current=0.2)
    src2 = Undulator()
    src = LightSource("test",src1,src2)

    for key in src1.keys():
        print(key,src1.to_dictionary()[key])

    for key in src2.keys():
        print(key,src2.to_dictionary()[key])


    print(src1.keys())
    print(src2.keys())

    print(src1.info())
    print(src2.info())


    print(src.info())


    src2.set_value_from_key_name("K_horizontal",33)

    print(src.info())
    assert (33,src2.get_value_from_key_name("K_horizontal"))

    src1.to_json("tmp_src1.json")
    src2.to_json("tmp_src2.json")



    print("==================== BeamLine elements: ==================")

    #
    # ideal elements
    #

    screen1 = Screen("screen1")

    screen1.keys()
    print(screen1.info())
    print(screen1.to_json())


    lens1 = IdealLens("lens1",3.0)

    lens1.keys()
    print(lens1.info())
    print(lens1.to_json())



    #
    # absorbers
    #
    filter1 = Filter("filter1","H2O",3.0e-6)

    filter1.keys()
    print(filter1.info())
    print(filter1.to_json())
    # print(filter1._support_dictionary)

    #slit1 = Slit.initialize_as_rectangle("slit1",2e-3,4e-3)
    slit1 = Slit(name="slit1",boundary_shape=Rectangle(-0.5e-3,0.5e-3,-2e-3,2e-3))

    slit1.keys()
    print(slit1.info())
    print(slit1.to_json())

    slit2 = Slit(name="slit2")

    slit2.set_rectangle(width=3e-4,height=5e-4)
    slit2.set_circle(radius=3e-4)
    print(slit2.info())
    print(slit2.to_json())


    stopper1 = BeamStopper(name="stopper1",boundary_shape=Rectangle(-0.5e-3,0.5e-3,-2e-3,2e-3))

    stopper1.keys()
    print(stopper1.info())
    print(stopper1.to_json())

    stopper2 = BeamStopper(name="stopper2")

    stopper2.set_rectangle(width=3e-4,height=5e-4)
    stopper2.set_circle(radius=3e-4)
    print(stopper2.info())
    print(stopper2.to_json())

    #
    # elements with shape: mirror, gratings, crystals
    #

    mirror1 = Mirror(name="mirror1")

    mirror1.keys()
    print(mirror1.info())
    print(mirror1.to_json())


    crystal1 = Crystal(name="crystal1")

    crystal1.keys()
    print(crystal1.info())
    print(crystal1.to_json())


    grating1 = Grating(name="grating1")

    grating1.keys()
    print(grating1.info())
    print(grating1.to_json())


    print("==================== BeamLine: ==================")

    beamline1 = Beamline()
    beamline1.set_light_source(src)
    beamline1.append_beamline_element(BeamlineElement(screen1,coordinates=ElementCoordinates(100.0)))

    print(beamline1.info())

    beamline = Beamline()

    mirror = Mirror(name="mirror1",
                    boundary_shape=Rectangle(x_left=-0.1,
                                             x_right=0.1,
                                             y_bottom=-0.6,
                                             y_top=0.6),
                    surface_shape=SphericalCylinder(radius=2500.0,
                                                    convexity=Convexity.UPWARD,
                                                    cylinder_direction=Direction.TANGENTIAL),
                    coating="Pt",
                    coating_thickness=1e-4)

    mirror_coordinates = ElementCoordinates(p=2.5,
                                            q=10.0,
                                            angle_radial=89.828,
                                            angle_azimuthal=180.0)

    beamline.append_beamline_element(beamline_element=BeamlineElement(optical_element=mirror,
                                                                      coordinates=mirror_coordinates))

    print(beamline.to_dictionary())
    print(beamline.to_full_dictionary())


"""
Represents a beamline.

BeamlineComponents can be attached at positions(BeamlinePosition), i.e. longitudinal, off-axis and inclined.
We can iterate of the components, find their positions or look for a specific component.
"""

from syned.syned_object import SynedObject
from syned.storage_ring.light_source import LightSource
from syned.storage_ring.empty_light_source import EmptyLightSource
from syned.beamline.beamline_element import BeamlineElement

from collections import OrderedDict

class Beamline(SynedObject):
    def __init__(self, light_source=LightSource(), beamline_elements_list=None):
        self._light_source = light_source
        if beamline_elements_list is None:
            self._beamline_elements_list = []
        else:
            self._beamline_elements_list = beamline_elements_list

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("light_source",       "Light Source",      ""),
                    ("beamline_elements_list",  "Beamline Elements", ""),
            ] )


    # overwrites the SynedObject method for dealing with list
    def to_dictionary(self):
        dict_to_save = OrderedDict()
        dict_to_save.update({"CLASS_NAME":self.__class__.__name__})

        dict_to_save["light_source"] = self._light_source.to_dictionary()
        dict_to_save["beamline_elements_list"] = [el.to_dictionary() for el in self._beamline_elements_list]

        return dict_to_save



    def get_light_source(self):
        return self._light_source

    def get_beamline_elements(self):
        return self._beamline_elements_list

    def set_light_source(self, light_source=LightSource()):
        if not (isinstance(light_source,LightSource) or isinstance(light_source,EmptyLightSource)):
            raise Exception("Input class must be of type: "+LightSource.__name__+" or "+EmptyLightSource.__name__)
        else:
            self._light_source = light_source

    def append_beamline_element(self, beamline_element=BeamlineElement()):
        if not isinstance(beamline_element,BeamlineElement):
            raise Exception("Input class must be of type: "+BeamlineElement.__name__)
        else:
            self._beamline_elements_list.append(beamline_element)

    def get_beamline_elements_number(self):
        return len(self._beamline_elements_list)

    def get_beamline_element_at(self, index):
        if index >= len(self._beamline_elements_list):
            raise IndexError("Index " + str(index) + " out of bounds")

        return self._beamline_elements_list[index]

    def duplicate(self):
        beamline_elements_list = []
        for beamline_element in self._beamline_elements_list:
            beamline_elements_list.append(beamline_element)

        return Beamline(light_source=self._light_source,
                        beamline_elements_list = beamline_elements_list)



"""
Base class for LighSource, which contains:
    - a name
    - an electron beam
    - a magnetic structure

"""
from syned.syned_object import SynedObject


class EmptyLightSource(SynedObject):
    def __init__(self, name="Empty"):
        self._name               = name
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",              "Name",""),
            ] )


    def get_name(self):
        return self._name




"""
Base class for LighSource, which contains:
    - a name
    - an electron beam
    - a magnetic structure

"""
from syned.syned_object import SynedObject
from syned.storage_ring.magnetic_structure import MagneticStructure
from syned.storage_ring.electron_beam import ElectronBeam

class LightSource(SynedObject):
    def __init__(self, name="Undefined", electron_beam=None, magnetic_structure=None):
        self._name = name
        if electron_beam is None:
            self._electron_beam = ElectronBeam()
        else:
            self._electron_beam      = electron_beam
        if magnetic_structure is None:
            self._magnetic_structure = MagneticStructure()
        else:
            self._magnetic_structure = magnetic_structure
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",              "Name",""),
                    ("electron_beam",     "Electron Beam",""),
                    ("magnetic_structure","Magnetic Strtructure",""),
            ] )


    def get_name(self):
        return self._name

    def get_electron_beam(self):
        return self._electron_beam

    def get_magnetic_structure(self):
        return self._magnetic_structure




if __name__ == "__main__":

    from syned.storage_ring.magnetic_structures.undulator import Undulator

    eb = ElectronBeam.initialize_as_pencil_beam( energy_in_GeV=2.0,energy_spread=0.0,current=0.5)
    ms = Undulator.initialize_as_vertical_undulator( K=1.8, period_length=0.038, periods_number=56.0 )

    light_source = LightSource(name="",electron_beam=eb,magnetic_structure=ms)

    print(light_source.info())





from syned.util.json_tools import load_from_json_file

from syned.storage_ring.electron_beam import ElectronBeam
from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.beamline.optical_elements.ideal_elements.screen import Screen
from syned.beamline.optical_elements.absorbers.slit import Slit

from syned.storage_ring.light_source import LightSource

from syned.beamline.beamline import Beamline
from syned.beamline.beamline_element import BeamlineElement
from syned.beamline.element_coordinates import ElementCoordinates

from syned.beamline.shape import MultiplePatch

if __name__ == "__main__":



    src1 = ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=6.0,current=0.2)
    src2 = Undulator()
    screen1 = Screen("screen1")

    patches = MultiplePatch()
    patches.append_rectangle(-0.02,-0.01,-0.001,0.001)
    patches.append_rectangle(0.01,0.02,-0.001,0.001)

    slit1 = Slit(name="slit1",boundary_shape=patches)

    mylist = [src1,src2,slit1]

    #
    # test individual elements
    #

    for i,element in enumerate(mylist):
        element.to_json("tmp_%d.json"%i)

    for i,element in enumerate(mylist):
        print("loading element %d"%i)
        tmp = load_from_json_file("tmp_%d.json"%i)
        print("returned class: ",type(tmp))



    #
    # test Ligtsource
    #
    lightsource1 = LightSource("test_source",src1,src2)
    lightsource1.to_json("tmp_100.json")

    tmp = load_from_json_file("tmp_100.json")
    print("returned class: ",type(tmp))
    print("\n-----------Info on: \n",tmp.info(),"----------------\n\n")

    print( tmp.get_electron_beam().info() )
    print( tmp.get_magnetic_structure().info() )

    #
    # test full beamline
    #

    SLIT1       = BeamlineElement(slit1,        coordinates=ElementCoordinates(p=15.0))

    MyList = [SLIT1]


    #
    # test BeamlineElement
    #


    for i,element in enumerate(MyList):

        element.to_json("tmp_%d.json"%(100+i))
        tmp = load_from_json_file("tmp_%d.json"%(100+i))
        print("returned class: ",type(tmp))
        print("\n-----------Info on: \n",tmp.info(),"----------------\n\n")

    #
    # test Beamline
    #

    BL = Beamline(LightSource(name="test",electron_beam=src1,magnetic_structure=src2),
                  [SLIT1])

    BL.to_json("tmp_200.json")

    #
    tmp = load_from_json_file("tmp_200.json")
    print("returned class: ",type(tmp))


    print(tmp.get_light_source().info())
    for element in tmp.get_beamline_elements():
        print("list element class: ",type(element))
        print(element.info())
    #
    #
    print("\n-----------Info on: \n",tmp.info(),"----------------\n\n")


"""
Implement an undulator with vertical and horizontal magnetic fields.
"""
import numpy
import scipy.constants as codata

cte = codata.e/(2*numpy.pi*codata.electron_mass*codata.c)

from syned.storage_ring.magnetic_structures.insertion_device import InsertionDevice

class Undulator(InsertionDevice):

    def __init__(self,
                 K_vertical = 0.0,
                 K_horizontal = 0.0,
                 period_length = 0.0,
                 number_of_periods = 1.0):
        InsertionDevice.__init__(self, K_vertical, K_horizontal, period_length, number_of_periods)

    def resonance_wavelength(self, gamma, theta_x=0.0, theta_z=0.0, harmonic=1.0):
        wavelength = (self.period_length() / (2.0*gamma **2)) * \
                     (1 + self.K_vertical()**2 / 2.0 + self.K_horizontal()**2 / 2.0 + \
                      gamma**2 * (theta_x**2 + theta_z ** 2))

        return wavelength/harmonic

    def resonance_frequency(self, gamma, theta_x=0.0, theta_z=0.0, harmonic=1.0):
        frequency = codata.c / self.resonance_wavelength(gamma, theta_x, theta_z)

        return frequency*harmonic

    def resonance_energy(self, gamma, theta_x=0.0, theta_z=0.0, harmonic=1.0):
        energy_in_ev = codata.h * self.resonance_frequency(gamma, theta_x, theta_z) / codata.e

        return energy_in_ev*harmonic

    def gaussian_central_cone_aperture(self, gamma, n=1.0):
        return (1/gamma)*numpy.sqrt((1.0/(2.0*n*self.number_of_periods())) * (1.0 + self.K_horizontal()**2/2.0 + self.K_vertical()**2/2.0))

    @classmethod
    def initialize_as_vertical_undulator(cls, K = 0.0, period_length = 0.0, periods_number = 1.0):
        return Undulator(K_vertical=K,
                         K_horizontal=0.0,
                         period_length=period_length,
                         number_of_periods=periods_number)
    #
    #
    #

    def get_sigmas_radiation(self, gamma, harmonic=1.0):
        # calculate sizes of the photon undulator beam
        # see formulas 25 & 30 in Elleaume (Onuki & Elleaume)
        photon_energy = self.resonance_energy(gamma,harmonic=harmonic)
        lambdan = 1e-10 * codata.h * codata.c / codata.e * 1e10 / photon_energy  # in m
        sigma_r = 2.740 / 4 / numpy.pi * numpy.sqrt(lambdan * self.number_of_periods() * self.period_length())
        sigma_r_prime =  0.69 * numpy.sqrt(lambdan / (self.length()))
        return sigma_r, sigma_r_prime

    def get_resonance_ring(self, gamma, harmonic=1.0, ring_order=1):
        K_value = numpy.sqrt( self.K_vertical()**2 + self.K_horizontal()**2)
        return 1.0/gamma*numpy.sqrt( ring_order / harmonic * (1+0.5*K_value**2) )


    def undulator_full_emitted_power(self, gamma, ring_current):
        """
        Reference: Eq. 18 in https://xdb.lbl.gov/Section2/Sec_2-1.html
        :param gamma:
        :param ring_current:
        :return:
        """
        ptot = (self.number_of_periods()/6) * codata.value('characteristic impedance of vacuum') * \
               ring_current * codata.e * 2 * numpy.pi * codata.c * gamma**2 * \
               (self.K_vertical()**2+self.K_horizontal()**2) / self.period_length()
        return ptot

    def get_photon_sizes_and_divergences(self,syned_electron_beam,harmonic=1):
        sr,srp = self.get_sigmas_radiation(syned_electron_beam.gamma(),harmonic=harmonic)
        sx,sz,sxp,szp = syned_electron_beam.get_sigmas_all()

        Sx = numpy.sqrt( sx**2 + sr**2)
        Sz = numpy.sqrt( sz**2 + sr**2)
        Sxp = numpy.sqrt( sxp**2 + srp**2)
        Szp = numpy.sqrt( szp**2 + srp**2)

        return Sx,Sz,Sxp,Szp

    def get_K_from_photon_energy(self,photon_energy,gamma,harmonic=1):
        m2ev = codata.c * codata.h / codata.e
        wavelength = harmonic * m2ev / photon_energy
        return numpy.sqrt(2*(((wavelength*2*gamma**2)/self.period_length())-1))

    def approximated_coherent_fraction_horizontal(self,syned_electron_beam,harmonic=1):
        Sx,Sy,Sxp,Syp = self.get_photon_sizes_and_divergences(syned_electron_beam,harmonic=harmonic)
        srad,sradp = self.get_sigmas_radiation(syned_electron_beam.gamma(),harmonic=harmonic)
        return srad * sradp / ( Sx * Sxp)

    def approximated_coherent_fraction_vertical(self,syned_electron_beam,harmonic=1):
        Sx,Sy,Sxp,Syp = self.get_photon_sizes_and_divergences(syned_electron_beam,harmonic=harmonic)
        srad,sradp = self.get_sigmas_radiation(syned_electron_beam.gamma(),harmonic=harmonic)
        return srad * sradp / ( Sy * Syp)

    def approximated_coherent_fraction(self,syned_electron_beam,harmonic=1):
        return self.approximated_coherent_fraction_horizontal(syned_electron_beam,harmonic=harmonic) * \
               self.approximated_coherent_fraction_vertical(syned_electron_beam, harmonic=harmonic)


if __name__ == "__main__":

    a = Undulator(number_of_periods=61.5, period_length=0.057)
    a.set_K_vertical_from_magnetic_field(0.187782)

    print(a._K_vertical)

    print (a.resonance_energy(gamma=5870.8540997356595))

    fd = a.to_full_dictionary()
    dict = a.to_dictionary()

    print(dict)

    for key in fd:
        print(key,fd[key][0])

    for key in fd:
        print(key,dict[key])

    print(a.keys())
    print(a.info())


    print("####### derived quantities ##########")

    from syned.storage_ring.electron_beam import ElectronBeam
    ebeam = ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=2.0,energy_spread=0.0,current=0.5)

    sigmas_radiation = a.get_sigmas_radiation(ebeam.gamma(),harmonic=1.0)
    print("sigmas radiation [m rad]:",sigmas_radiation)


    ring = a.get_resonance_ring(ebeam.gamma(), harmonic=1.0, ring_order=1)
    print("first ring at [rad]:",ring)

    pow = a.undulator_full_emitted_power(ebeam.gamma(), ebeam.current())
    print("Total emission [W]",pow)

    sizes = a.get_photon_sizes_and_divergences(ebeam,harmonic=1)
    print("Sizes: ",sizes)

    print("Resonance: ",a.resonance_energy(ebeam.gamma()))

    print("K at 444 eV",a.get_K_from_photon_energy(444.0,ebeam.gamma()))
    print("K at 200 eV", a.get_K_from_photon_energy(200.0, ebeam.gamma()))

    print("CF H V HV: ",
          a.approximated_coherent_fraction_horizontal(ebeam,harmonic=1),
          a.approximated_coherent_fraction_vertical(ebeam, harmonic=1),
          a.approximated_coherent_fraction(ebeam, harmonic=1),)



import numpy
from syned.syned_object import SynedObject
from collections import OrderedDict

class Convexity:
    NONE = -1
    UPWARD = 0
    DOWNWARD = 1

class Direction:
    TANGENTIAL = 0
    SAGITTAL = 1

class Side:
    SOURCE = 0
    IMAGE = 1
#
# main shape subclasses:
#      SurfaceShape to caracterize the shape (sphere etc.) of the optical element surface
#      BoundaryShape to characterize the optical element dimensions (rectangle, etc.)
#
class Shape(SynedObject):
    def __init__(self):
        SynedObject.__init__(self)

class SurfaceShape(Shape):
    def __init__(self, convexity = Convexity.UPWARD):
        Shape.__init__(self)

        self._convexity = convexity

class BoundaryShape(Shape):
    def __init__(self):
        Shape.__init__(self)
        
    def get_boundaries(self):
        raise NotImplementedError()

#
# Subclasses for SurfaceShape
#

class Cylinder:
    def __init__(self, cylinder_direction=Direction.TANGENTIAL):
        self._cylinder_direction = cylinder_direction

class Conic(SurfaceShape):
    def __init__(self, 
                 conic_coefficients=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], 
                 convexity=Convexity.UPWARD):
        SurfaceShape.__init__(self, convexity)

        self._conic_coefficients = conic_coefficients

class Plane(SurfaceShape):
    def __init__(self):
        SurfaceShape.__init__(self, convexity=Convexity.NONE)

class Sphere(SurfaceShape):
    def __init__(self, radius=1.0, convexity=Convexity.UPWARD):
        SurfaceShape.__init__(self, convexity=Convexity.NONE)
        self._radius = radius

    def get_radius(self):
        return self._radius

    def initialize_from_p_q(self, p=2.0, q=1.0, grazing_angle=0.003):
        self.self._radius = Sphere.get_radius_from_p_q(p, q, grazing_angle)

    @classmethod
    def get_radius_from_p_q(cls, p=2.0, q=1.0, grazing_angle=0.003):
        # 1/p + 1/q = 2/(R cos(pi/2 - gr.a.))
        return (2*p*q/(p+q))/numpy.sin(grazing_angle)

class SphericalCylinder(Sphere, Cylinder):
    def __init__(self, 
                 radius=1.0, 
                 convexity=Convexity.UPWARD, 
                 cylinder_direction=Direction.TANGENTIAL):
        Sphere.__init__(self, radius, convexity)
        Cylinder.__init__(self, cylinder_direction)

    def initialize_from_p_q(self, p=2.0, q=1.0, grazing_angle=0.003):
        if self._cylinder_direction == Direction.TANGENTIAL:
            self._conic_coefficients[9] = -Sphere.get_radius_from_p_q(p, q, grazing_angle)**2
        elif self._cylinder_direction == Direction.SAGITTAL:
            self._conic_coefficients[9] = -SphericalCylinder.get_radius_from_p_q_sagittal(p, q, grazing_angle)**2

    @classmethod
    def get_radius_from_p_q_sagittal(cls, p=2.0, q=1.0, grazing_angle=0.003):
        # 1/p + 1/q = 2 cos(pi/2 - gr.a.)/r
        return (2*p*q/(p+q))*numpy.sin(grazing_angle)

class Ellipsoid(SurfaceShape):
    def __init__(self, min_axis=0.0, maj_axis=0.0, convexity=Convexity.UPWARD):
        SurfaceShape.__init__(self, convexity)

        self._min_axis = min_axis
        self._maj_axis = maj_axis

    def initialize_from_p_q(self, p=2.0, q=1.0, grazing_angle=0.003):
        self._min_axis, self._maj_axis = Ellipsoid.get_axis_from_p_q(p, q, grazing_angle)

    def get_p_q(self, grazing_angle=0.003):
        return Ellipsoid.get_p_q_from_axis(self._min_axis, self._maj_axis, grazing_angle)

    @classmethod
    def get_axis_from_p_q(cls, p=2.0, q=1.0, grazing_angle=0.003):
        # see calculation of ellipse axis in shadow_kernel.f90 row 3605
        min_axis = 2*numpy.sqrt(p*q)*numpy.sin(grazing_angle)
        maj_axis = (p + q)

        return min_axis, maj_axis

    @classmethod
    def get_p_q_from_axis(cls, min_axis=2.0, maj_axis=1.0, grazing_angle=0.003):
        a = 0.5*maj_axis
        b = 0.5*min_axis
        p = a + numpy.sqrt(a**2 - (b/numpy.sin(grazing_angle))**2)
        q = 2*a - p

        return p, q

class EllipticalCylinder(Ellipsoid, Cylinder):
    def __init__(self, 
                 min_axis=0.0, 
                 maj_axis=0.0, 
                 convexity=Convexity.UPWARD, 
                 cylinder_direction=Direction.TANGENTIAL):
        Ellipsoid.__init__(self, min_axis, maj_axis, convexity)
        Cylinder.__init__(self, cylinder_direction)

    def initialize_from_p_q(self, p=2.0, q=1.0, grazing_angle=0.003):
        if self._cylinder_direction == Direction.SAGITTAL:
            raise NotImplementedError("Operation not possible for SAGITTAL direction")

        return super().initialize_from_p_q(p, q, grazing_angle)

    def get_p_q(self, grazing_angle=0.003):
        if self._cylinder_direction == Direction.SAGITTAL:
            raise NotImplementedError("Operation not possible for SAGITTAL direction")

        return super().get_p_q(grazing_angle)

class Paraboloid(SurfaceShape):
    def __init__(self, 
                 parabola_parameter=0.0, 
                 convexity=Convexity.UPWARD):
        SurfaceShape.__init__(self, convexity)

        self._parabola_parameter = parabola_parameter

    def initialize_from_p_q(self, p=2.0, q=1.0, grazing_angle=0.003, at_infinity=Side.SOURCE):
        self._parabola_parameter = Paraboloid.get_parabola_parameter_from_p_q(p, q, grazing_angle, at_infinity)

    @classmethod
    def get_parabola_parameter_from_p_q(cls, p=2.0, q=1.0, grazing_angle=0.003, at_infinity=Side.SOURCE):
        if at_infinity == Side.IMAGE:
            return 2*p*(numpy.sin(grazing_angle))**2
        elif at_infinity == Side.SOURCE:
            return 2*q*(numpy.sin(grazing_angle))**2


class ParabolicCylinder(Paraboloid, Cylinder):
    def __init__(self, 
                 parabola_parameter=0.0, 
                 convexity=Convexity.UPWARD, 
                 cylinder_direction=Direction.TANGENTIAL):
        Paraboloid.__init__(self, parabola_parameter, convexity)
        Cylinder.__init__(self, cylinder_direction)

    def initialize_from_p_q(self, p=2.0, q=1.0, grazing_angle=0.003, at_infinity=Side.SOURCE):
        if self._cylinder_direction == Direction.SAGITTAL:
            raise NotImplementedError("Operation not possible for SAGITTAL direction")

        return super().initialize_from_p_q(p, q, grazing_angle, at_infinity)

class Hyperboloid(SurfaceShape):
    def __init__(self, min_axis=0.0, maj_axis=0.0, convexity=Convexity.UPWARD):
        SurfaceShape.__init__(self, convexity)

        self._min_axis = min_axis
        self._maj_axis = maj_axis

    def initialize_from_p_q(self, p=2.0, q=1.0, grazing_angle=0.003):
        raise NotImplementedError("TBD")

    def get_p_q(self, grazing_angle=0.003):
        raise NotImplementedError("TBD")

class HyperbolicCylinder(Hyperboloid, Cylinder):
    def __init__(self, 
                 min_axis=0.0, 
                 maj_axis=0.0, 
                 convexity=Convexity.UPWARD, 
                 cylinder_direction=Direction.TANGENTIAL):
        Hyperboloid.__init__(self, min_axis, maj_axis, convexity)
        Cylinder.__init__(self, cylinder_direction)

class Toroidal(SurfaceShape):
    def __init__(self, min_radius=0.0, maj_radius=0.0):
        SurfaceShape.__init__(self, convexity=Convexity.NONE)
        
        self._min_radius = min_radius
        self._maj_radius = maj_radius

    def initialize_from_p_q(self, p=2.0, q=1.0, grazing_angle=0.003):
        self._maj_radius = Sphere.get_radius_from_p_q(p, q, grazing_angle)
        self._min_radius = SphericalCylinder.get_radius_from_p_q_sagittal(p, q, grazing_angle)

        # FROM SHADOW3:
        #! C
        #! C NOTE : The major radius is the in reality the radius of the torus
        #! C max. circle. The true major radius is then
        #! C
        #        R_MAJ	=   R_MAJ - R_MIN
        self._maj_radius -= self._min_radius

class NumbericalMesh(SurfaceShape):
    def __init__(self):
        SurfaceShape.__init__(self, convexity=Convexity.NONE)


#
# subclasses for BoundaryShape
#


class Rectangle(BoundaryShape):
    def __init__(self, x_left=-0.010, x_right=0.010, y_bottom=-0.020, y_top=0.020):
        super().__init__()

        self._x_left   = x_left
        self._x_right  = x_right
        self._y_bottom = y_bottom
        self._y_top    = y_top

        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("x_left"          , "x (width) minimum (signed)   ", "m" ),
                    ("x_right"         , "x (width) maximum (signed)   ", "m" ),
                    ("y_bottom"        , "y (length) minimum (signed)  ", "m" ),
                    ("y_top"           , "y (length) maximum (signed)  ", "m" ),
            ] )

    def get_boundaries(self):
        return self._x_left, self._x_right, self._y_bottom, self._y_top

    def set_boundaries(self,x_left=-0.010, x_right=0.010, y_bottom=-0.020, y_top=0.020):
        self._x_left = x_left
        self._x_right = x_right
        self._y_bottom = y_bottom
        self._y_top = y_top

    def set_width_and_length(self,width=10e-3,length=30e-3):
        self._x_left = -0.5 * width
        self._x_right = 0.5 * width
        self._y_bottom = -0.5 * length
        self._y_top = 0.5 * length

class Ellipse(BoundaryShape):
    def __init__(self, a_axis_min=-10e-6, a_axis_max=10e-6, b_axis_min=-5e-6, b_axis_max=5e-6):
        super().__init__()

        self._a_axis_min   = a_axis_min
        self._a_axis_max  = a_axis_max
        self._b_axis_min = b_axis_min
        self._b_axis_max    = b_axis_max
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("a_axis_min"         , "x (width) axis starts (signed)  ", "m" ),
                    ("a_axis_max"        , "x (width) axis ends (signed)    ", "m" ),
                    ("b_axis_min"       , "y (length) axis starts (signed) ", "m" ),
                    ("b_axis_max"          , "y (length) axis ends (signed)   ", "m" ),
            ] )

    def get_boundaries(self):
        return self._a_axis_min, self._a_axis_max, self._b_axis_min, self._b_axis_max

    def get_axis(self):
        return numpy.abs(self._a_axis_max - self._a_axis_min), numpy.abs(self._b_axis_max - self._b_axis_min)

class Circle(BoundaryShape):
    def __init__(self,radius=50e-6,x_center=0.0,y_center=0.0):
        super().__init__()

        self._radius = radius
        self._x_center = x_center
        self._y_center = y_center
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("radius"              , "radius  ", "m" ),
                    ("x_center"            , "x center (signed)    ", "m" ),
                    ("y_center"            , "y center (signed)    ", "m" ),
            ] )

    def get_boundaries(self):
        return self._radius, self._x_center, self._y_center

    def set_boundaries(self, radius=1.0, x_center=0.0, y_center=0.0):
        self._radius = radius
        self._x_center = x_center
        self._y_center = y_center

    def get_radius(self):
        return self._radius

    def get_center(self):
        return [self._x_center,self._y_center]


class MultiplePatch(BoundaryShape):
    def __init__(self, patch_list=None):
        super().__init__()

        if patch_list is None:
            self._patch_list = []
        else:
            self._patch_list = patch_list
        self._set_support_text([
                    ("multiple_patch_list",  "Multiple Patch", ""),
            ])


    # overwrites the SynedObject method for dealing with list
    def to_dictionary(self):
        dict_to_save = OrderedDict()
        dict_to_save.update({"CLASS_NAME":self.__class__.__name__})

        dict_to_save["multiple_patch_list"] = [el.to_dictionary() for el in self._patch_list]

        return dict_to_save


    def reset(self):
        self._patch_list = []

    def get_number_of_patches(self):
        return len(self._patch_list)

    def get_boundaries(self):
        boundaries_list = []
        for i in range(self.get_number_of_patches()):
            boundaries_list.extend(list(self._patch_list[i].get_boundaries()))
        return tuple(boundaries_list)

    def append_patch(self,patch=BoundaryShape()):
        self._patch_list.append(patch)

    def append_rectangle(self,x_left=-0.010,x_right=0.010,y_bottom=-0.020,y_top=0.020):
        self.append_patch(Rectangle(x_left=x_left, x_right=x_right, y_bottom=y_bottom, y_top=y_top))

    def append_circle(self,radius, x_center=0.0, y_center=0.0):
        self.append_patch(Circle(radius, x_center=x_center, y_center=y_center))

    def append_ellipse(self,a_axis_min, a_axis_max, b_axis_min, b_axis_max):
        self.append_patch(Ellipse(a_axis_min, a_axis_max, b_axis_min, b_axis_max))

    def get_patches(self):
        return self._patch_list

    def get_patch(self,index):
        return self.get_patches()[index]

    def get_name_of_patch(self,index):
        return self._patch_list[index].__class__.__name__

class DoubleRectangle(MultiplePatch):
    def __init__(self, x_left1=-0.010, x_right1=0.0, y_bottom1=-0.020, y_top1=0.0,
                        x_left2=-0.010, x_right2=0.010, y_bottom2=-0.001, y_top2=0.020):
        super().__init__()
        self.reset()
        self.append_patch(Rectangle(x_left=x_left1, x_right=x_right1, y_bottom=y_bottom1, y_top=y_top1))
        self.append_patch(Rectangle(x_left=x_left2, x_right=x_right2, y_bottom=y_bottom2, y_top=y_top2))

    def set_boundaries(self,x_left1=-0.010, x_right1=0.0, y_bottom1=-0.020, y_top1=0.0,
                        x_left2=-0.010, x_right2=0.010, y_bottom2=-0.001, y_top2=0.020):
        self._patch_list[0].set_boundaries(x_left1, x_right1, y_bottom1, y_top1)
        self._patch_list[1].set_boundaries(x_left2, x_right2, y_bottom2, y_top2)

class DoubleEllipse(MultiplePatch):
    def __init__(self, a_axis_min1=-0.010, a_axis_max1=0.0,   b_axis_min1=-0.020, b_axis_max1=0.0,
                       a_axis_min2=-0.010, a_axis_max2=0.010, b_axis_min2=-0.001, b_axis_max2=0.020):
        super().__init__()
        self.reset()
        self.append_patch(Ellipse(a_axis_min1, a_axis_max1, b_axis_min1, b_axis_max1))
        self.append_patch(Ellipse(a_axis_min2, a_axis_max2, b_axis_min2, b_axis_max2))

    def set_boundaries(self,a_axis_min1=-0.010, a_axis_max1=0.0,   b_axis_min1=-0.020, b_axis_max1=0.0,
                            a_axis_min2=-0.010, a_axis_max2=0.010, b_axis_min2=-0.001, b_axis_max2=0.020):
        self._patch_list[0].set_boundaries(a_axis_min1,a_axis_max1,b_axis_min1,b_axis_max1)
        self._patch_list[1].set_boundaries(a_axis_min2,a_axis_max2,b_axis_min2,b_axis_max2)

class DoubleCircle(MultiplePatch):
    def __init__(self, radius1=50e-6,x_center1=0.0,y_center1=0.0,
                       radius2=50e-6,x_center2=100e-6,y_center2=100e-6):
        super().__init__()
        self.reset()
        self.append_patch(Circle(radius1,x_center1,y_center1))
        self.append_patch(Circle(radius2,x_center2,y_center2))

    def set_boundaries(self,radius1=50e-6,x_center1=0.0,y_center1=0.0,
                            radius2=50e-6,x_center2=100e-6,y_center2=100e-6):
        self._patch_list[0].set_boundaries(radius1,x_center1,y_center1)
        self._patch_list[1].set_boundaries(radius2,x_center2,y_center2)



if __name__=="__main__":

    # ell = Ellipsoid()
    # ell.initialize_from_p_q(20, 10, 0.2618)
    #
    # print (ell._min_axis/2, ell._maj_axis/2)
    #
    # ell = Ellipsoid(min_axis=ell._min_axis, maj_axis=ell._maj_axis)
    #
    # print(ell.get_p_q(0.2618))


    # circle = Circle(3.0)
    #
    # print(circle.get_radius(),circle.get_center())
    # print(circle.get_boundaries())




    # patches = MultiplePatch()
    #
    # patches.append_rectangle(-0.02,-0.01,-0.001,0.001)
    # patches.append_rectangle(0.01,0.02,-0.001,0.001)
    #
    # print(patches.get_number_of_patches(),patches.get_boundaries())
    #
    # for patch in patches.get_patches():
    #     print(patch.info())
    #
    # print("Patch 0 is: ",patches.get_name_of_patch(0))
    # print("Patch 1 is: ",patches.get_name_of_patch(1))
    # print(patches.get_boundaries())


    double_rectangle = DoubleRectangle()
    double_rectangle.set_boundaries(-0.02,-0.01,-0.001,0.001,0.01,0.02,-0.001,0.001)
    print("Rectangle 0 is: ",double_rectangle.get_name_of_patch(0))
    print("Rectangle 1 is: ",double_rectangle.get_name_of_patch(1))
    print(double_rectangle.get_boundaries())



"""
Base class for all beamline components.
Enforce to name every every component.
Every beamline component can store settings
"""

from syned.syned_object import SynedObject
from syned.beamline.optical_element import OpticalElement
from syned.beamline.element_coordinates import ElementCoordinates

class BeamlineElement(SynedObject):
    def __init__(self, optical_element=OpticalElement(), coordinates=ElementCoordinates()):
        self._optical_element = optical_element
        self._coordinates = coordinates
        # support text containg name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("optical_element",       "Optical Element",      ""),
                    ("coordinates",           "Element coordinates",  ""),
            ] )

    def get_optical_element(self):
        return self._optical_element

    def get_coordinates(self):
        return self._coordinates


"""
Represents an ideal lens.
"""
from syned.beamline.optical_elements.ideal_elements.ideal_element import IdealElement

class Screen(IdealElement):
    def __init__(self, name="Undefined"):
        IdealElement.__init__(self, name=name)



# THIS FILE IS GENERATED FROM syned SETUP.PY
short_version = '1.0.17'
version = '1.0.17'
full_version = '1.0.17.dev0+ac41b0e'
git_revision = 'ac41b0eaeb5546558a764f6d9c4b3ea95b52f3d7'
release = False

if not release:
    version = full_version
    short_version += ".dev"




from syned.beamline.shape import BoundaryShape
from syned.beamline.shape import Rectangle, Ellipse

from syned.beamline.optical_elements.absorbers.absorber import Absorber

class BeamStopper(Absorber):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Absorber.__init__(self, name=name, boundary_shape=boundary_shape)

    def set_rectangle(self,width=3e-3,height=4e-3):
        self._boundary_shape=Rectangle(-0.5*width,0.5*width,-0.5*height,0.5*height)

    def set_circle(self,radius=3e-3):
        self._boundary_shape=Ellipse(-0.5*radius,0.5*radius,-0.5*radius,0.5*radius)


"""
Base class for a Bending Magnet


"""
from syned.storage_ring.magnetic_structure import MagneticStructure
import numpy
import scipy.constants as codata

class BendingMagnet(MagneticStructure):
    def __init__(self, radius=1.0, magnetic_field=1.0, length=1.0):
        """
        Constructor.
        :param radius: Physical Radius/curvature of the magnet in m
        :param magnetic_field: Magnetic field strength in T
        :param length: physical length of the bending magnet (along the arc) in m.
        """
        MagneticStructure.__init__(self)
        self._radius         = radius
        self._magnetic_field = magnetic_field
        self._length         = length

        # support text contaning name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("radius"          , "Radius of bending magnet" , "m"    ),
                    ("magnetic_field"  , "Magnetic field",            "T"    ),
                    ("length"          , "Bending magnet length",     "m"   ),
            ] )

    #
    #methods for practical calculations
    #
    @classmethod
    def initialize_from_magnetic_field_divergence_and_electron_energy(cls,magnetic_field=1.0,divergence=1e-3,electron_energy_in_GeV=1.0):
        """
        Constructor from  magnetic field divergence and electron energy
        :param magnetic_field: in T
        :param divergence: in rad
        :param electron_energy_in_GeV: in GeV
        :return:
        """
        magnetic_radius = cls.calculate_magnetic_radius(magnetic_field, electron_energy_in_GeV)
        return BendingMagnet(magnetic_radius,magnetic_field,numpy.abs(divergence * magnetic_radius))

    @classmethod
    def initialize_from_magnetic_radius_divergence_and_electron_energy(cls,magnetic_radius=10.0,divergence=1e-3,electron_energy_in_GeV=1.0):
        """
        Constructor from  magnetic radius, divergence and electron energy
        :param magnetic_radius: in m
        :param divergence: in rad
        :param electron_energy_in_GeV: in GeV
        :return:
        """
        magnetic_field = cls.calculate_magnetic_field(magnetic_radius, electron_energy_in_GeV)
        return BendingMagnet(magnetic_radius,magnetic_field,numpy.abs(divergence * magnetic_radius))


    def length(self):
        """
        return length in m
        :return:
        """
        return self._length

    def magnetic_field(self):
        """
        return magnetic field in T
        :return:
        """
        return self._magnetic_field

    def radius(self):
        """
        return radius in m
        :return:
        """
        return self._radius

    def horizontal_divergence(self):
        """
        return horizontal divergence in rad
        :return:
        """
        return numpy.abs(self.length()/self.radius())

    def get_magnetic_field(self, electron_energy_in_GeV):
        """
        calculates magnetic field from magnetic radius and electron energy
        :param electron_energy_in_GeV:
        :return:
        """
        return BendingMagnet.calculate_magnetic_field(self._radius, electron_energy_in_GeV)

    def get_magnetic_radius(self, electron_energy_in_GeV):
        """
        calculates magnetic radius from magnetic field and electron energy
        :param electron_energy_in_GeV:
        :return:
        """
        return BendingMagnet.calculate_magnetic_radius(self._magnetic_field, electron_energy_in_GeV)


    def get_critical_energy(self, electron_energy_in_GeV, method=1):
        """
        Calculates critical energy
        :param electron_energy_in_GeV:
        :param method: 0= uses magnetic radius, 1=uses magnetic field
        :return: Photon Critical energy in eV
        """

        if method == 0:
            return BendingMagnet.calculate_critical_energy(self._radius, electron_energy_in_GeV)
        else:
            return BendingMagnet.calculate_critical_energy_from_magnetic_field(self._magnetic_field, electron_energy_in_GeV)



    # for equations, see for example https://people.eecs.berkeley.edu/~attwood/srms/2007/Lec09.pdf
    @classmethod
    def calculate_magnetic_field(cls, magnetic_radius, electron_energy_in_GeV):
        """
        Calculates magnetic field
        :param magnetic_radius:
        :param electron_energy_in_GeV:
        :return: magnetic field in T
        """
        # return 3.334728*electron_energy_in_GeV/magnetic_radius
        return 1e9 / codata.c * electron_energy_in_GeV / magnetic_radius

    @classmethod
    def calculate_magnetic_radius(cls, magnetic_field, electron_energy_in_GeV):
        """
        Calculates magnetic radius
        :param magnetic_field:
        :param electron_energy_in_GeV:
        :return:
        """
        # return 3.334728*electron_energy_in_GeV/magnetic_field
        return 1e9 / codata.c * electron_energy_in_GeV / magnetic_field

    @classmethod
    def calculate_critical_energy(cls, magnetic_radius, electron_energy_in_GeV):
        """
        Calculates critical energy
        :param magnetic_radius:
        :param electron_energy_in_GeV:
        :return:
        """
        # omega = 3 g3 c / (2r)
        gamma = 1e9 * electron_energy_in_GeV / (codata.m_e *  codata.c**2 / codata.e)
        critical_energy_J = 3 * codata.c * codata.hbar * gamma**3 / (2 * numpy.abs(magnetic_radius))
        critical_energy_eV = critical_energy_J / codata.e
        return critical_energy_eV

    @classmethod
    def calculate_critical_energy_from_magnetic_field(cls, magnetic_field, electron_energy_in_GeV):
        """
        Calculates critical energy
        :param magnetic_field:
        :param electron_energy_in_GeV:
        :return:
        """
        # omega = 3 g3 c / (2r)
        magnetic_radius = cls.calculate_magnetic_radius(magnetic_field, electron_energy_in_GeV)
        return cls.calculate_critical_energy(magnetic_radius, electron_energy_in_GeV)


if __name__ == "__main__":
    print("input for ESRF: ")
    B = BendingMagnet.calculate_magnetic_field(25.0,6.04)
    print(">> B = ",B)
    print(">> R = ",BendingMagnet.calculate_magnetic_radius(B,6.04))
    print(">> Ec = ",BendingMagnet.calculate_critical_energy(25.0,6.04))
    print(">> Ec = ",BendingMagnet.calculate_critical_energy_from_magnetic_field(B, 6.04))
    BB = BendingMagnet.calculate_magnetic_radius (BendingMagnet.calculate_magnetic_radius (B,6.04),6.04)
    RR = BendingMagnet.calculate_magnetic_radius(BendingMagnet.calculate_magnetic_field(25.0,6.04), 6.04)
    assert(BB == B)
    assert(RR == 25.0)


    print("input for ALS: ")
    B = BendingMagnet.calculate_magnetic_field(5.0,1.9)
    print(">> B = ",B)
    print(">> R = ",BendingMagnet.calculate_magnetic_radius (B,1.9))
    print(">> Ec = ",BendingMagnet.calculate_critical_energy(5.0,1.9))
    print(">> Ec = ",BendingMagnet.calculate_critical_energy_from_magnetic_field(B, 1.9))
    BB = BendingMagnet.calculate_magnetic_radius (BendingMagnet.calculate_magnetic_radius (B,1.9),1.9)
    RR = BendingMagnet.calculate_magnetic_radius(BendingMagnet.calculate_magnetic_field(5.0, 1.9), 1.9)
    assert(BB == B)
    assert(RR == 5.0)